from __future__ import annotations

import re
from contextlib import ExitStack
from functools import wraps
from typing import TYPE_CHECKING, Sequence

import pytest
from nox import parametrize
from nox._decorators import Func
from nox._options import options as nox_options
from nox.manifest import Manifest

from noxopt import Annotated, NoxOpt, Option, Session


if TYPE_CHECKING:
    from typing import Protocol

    class Executor(Protocol):
        def __call__(
            self,
            session: str = "",
            tag: str = "",
            posargs: Sequence[str] = ...,
            raises: Exception | tuple[type[Exception], str | re.Pattern] | None = ...,
            prints: str | re.Pattern | None = ...,
            logs: str | re.Pattern | None = ...,
        ) -> None:
            ...


@pytest.fixture
def registry(monkeypatch: pytest.MonkeyPatch) -> dict[str, Func]:
    funcs: dict[str, Func] = {}
    monkeypatch.setattr("nox.registry._REGISTRY", funcs)
    return funcs


@pytest.fixture
def execute(
    registry: dict[str, Func],
    capsys: pytest.CaptureFixture,
    caplog: pytest.LogCaptureFixture,
) -> Executor:
    def execute(
        session: str = "",
        tag: str = "",
        posargs: Sequence[str] = (),
        raises: Exception | tuple[type[Exception], str | re.Pattern] | None = None,
        prints: str | re.Pattern | None = None,
        logs: str | re.Pattern | None = None,
    ) -> None:
        manifest = Manifest(
            {k: v for k, v in registry.items() if k == session or tag in v.tags},
            nox_options.namespace(posargs=posargs),
        )

        if raises:
            if isinstance(raises, tuple):
                etype, epat = raises
            else:
                etype, epat = type(raises), str(raises)

            if not isinstance(epat, re.Pattern):
                epat = re.compile(epat)
        else:
            etype = epat = None

        with ExitStack() as context:
            if etype and epat:
                context.enter_context(pytest.raises(etype, match=epat))
            for runner in manifest:
                runner.execute()

        if prints:
            if not isinstance(prints, re.Pattern):
                prints = re.compile(prints)

            readout = capsys.readouterr()
            for line in readout.out.split("\n") + readout.err.split("\n"):
                if prints.search(line):
                    break
            else:
                raise AssertionError(f"No output matches pattern {prints.pattern!r}")

        if logs:
            if not isinstance(logs, re.Pattern):
                logs = re.compile(logs)

            for line in caplog.messages:
                if logs.search(line):
                    break
            else:
                raise AssertionError(f"No log matches pattern {logs.pattern!r}")

    return execute


def test_session_with_simple_option(execute: Executor):
    group = NoxOpt()

    number_value = None

    @group.session
    def my_session(session: Session, number: int = 0) -> None:
        nonlocal number_value
        number_value = number

    execute(session="my-session")
    assert number_value == 0
    execute(session="my-session", posargs=["--number", "1"])
    assert number_value == 1


def test_session_with_annotated_option(execute: Executor):
    group = NoxOpt()

    number_value = None

    @group.session
    def my_session(
        session: Session,
        # double the given number
        number: Annotated[int, Option(type=lambda i: int(i) * 2)] = 0,
    ) -> None:
        nonlocal number_value
        number_value = number

    execute(session="my-session", posargs=["--number", "2"])
    assert number_value == 4


def test_all_options_are_flags(execute: Executor):
    group = NoxOpt()

    with pytest.raises(ValueError, match="only supports flags"):

        @group.session
        def my_session(
            session: Session, number: Annotated[int, Option(flags=["not-a-flag"])] = 0
        ) -> None:
            ...


def test_expect_metadata_to_be_option(execute: Executor):
    group = NoxOpt()

    with pytest.raises(ValueError, match="metadata must be an Option"):

        @group.session
        def my_session(
            session: Session, number: Annotated[int, "not an option"] = 0
        ) -> None:
            ...


def test_only_kwarg_params(execute: Executor):
    group = NoxOpt()

    with pytest.raises(TypeError, match="Found non-keyword session parameters"):

        @group.session
        def my_session(session: Session, *args: int) -> None:
            ...

    with pytest.raises(TypeError, match="Found non-keyword session parameters"):

        @group.session
        def my_session(session: Session, **args: int) -> None:
            ...


def test_no_extra_metadata(execute: Executor):
    group = NoxOpt()

    with pytest.raises(ValueError, match="has extra metadata"):

        @group.session
        def my_session(
            session: Session, number: Annotated[int, Option(), "extra-junk"] = 0
        ) -> None:
            ...


def tests_error_on_conflicting_options(execute: Executor):
    group = NoxOpt()

    @group.session
    def my_session(session: Session, some_option: int = 0) -> None:
        ...

    with pytest.raises(ValueError, match="Conflicting session options"):

        @group.session
        def my_session(session: Session, some_option: bool = 0) -> None:
            ...

    @group.session
    def my_session(
        session: Session,
        some_option: Annotated[int, Option(type=int)] = 0,
    ) -> None:
        ...

    with pytest.raises(ValueError, match="Conflicting session options"):

        @group.session
        def my_session(
            session: Session,
            some_option: Annotated[int, Option(type=int, help="different")] = 0,
        ) -> None:
            ...


def test_auto_tags_no_prefix(registry):
    group = NoxOpt(auto_tag=True)

    @group.session
    def a_x_1(session: Session) -> None:
        ...

    @group.session
    def a_x_2(session: Session) -> None:
        ...

    @group.session
    def a_y_1(session: Session) -> None:
        ...

    @group.session
    def a_y_2(session: Session) -> None:
        ...

    @group.session
    def b_x_1(session: Session) -> None:
        ...

    @group.session
    def b_x_2(session: Session) -> None:
        ...

    assert set(registry["a-x-1"].tags) == {"a", "a-x"}
    assert set(registry["a-x-2"].tags) == {"a", "a-x"}
    assert set(registry["a-y-1"].tags) == {"a", "a-y"}
    assert set(registry["a-y-2"].tags) == {"a", "a-y"}
    assert set(registry["b-x-1"].tags) == {"b-x"}
    assert set(registry["b-x-2"].tags) == {"b-x"}


def test_setup_funcs(execute: Executor):
    group = NoxOpt(auto_tag=True)

    calls = []

    @group.setup
    def setup_app(session: Session):
        calls.append("setup-all")

    @group.setup("session-x")
    def setup_x(session: Session, setup_param: str = "not-set") -> None:
        assert setup_param == "is-set"
        calls.append("setup-x")

    @group.setup("session-y")
    def setup_y(session: Session, setup_param: str = "not-set") -> None:
        assert setup_param == "is-set"
        calls.append("setup-y")

    @group.session
    def session_x_1(session: Session, session_param: bool = False):
        assert session_param
        calls.append("session-x-1")

    @group.session
    def session_x_2(session: Session, session_param: bool = False):
        assert session_param
        calls.append("session-x-2")

    @group.session
    def session_y_1(session: Session, session_param: bool = False):
        assert session_param
        calls.append("session-y-1")

    @group.session
    def session_y_2(session: Session, session_param: bool = False):
        assert session_param
        calls.append("session-y-2")

    execute(tag="session-x", posargs=["--setup-param", "is-set", "--session-param"])
    assert calls == [
        "setup-all",
        "setup-x",
        "session-x-1",
        "setup-all",
        "setup-x",
        "session-x-2",
    ]

    calls.clear()

    execute(tag="session-y", posargs=["--setup-param", "is-set", "--session-param"])
    assert calls == [
        "setup-all",
        "setup-y",
        "session-y-1",
        "setup-all",
        "setup-y",
        "session-y-2",
    ]


def test_required_argument(execute: Executor, caplog: pytest.LogCaptureFixture):
    group = NoxOpt()

    @group.session
    def session_with_required(session: Session, required: str):
        assert required == "expected"

    execute(
        "session-with-required",
        posargs=[],
        raises=SystemExit(2),
        prints=r"the following arguments are required: --required",
    )

    execute(
        "session-with-required",
        posargs=["--required", "wrong"],
        logs="AssertionError",
    )

    execute(
        "session-with-required",
        posargs=["--required", "expected"],
        logs="AssertionError",
    )


def test_functools_wraps(execute: Executor) -> None:
    group = NoxOpt()

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs) -> None:
            func(*args, extra_option=True, **kwargs)

        return wrapper

    @group.session
    @decorator
    def my_session(
        session: Session,
        my_option: bool = False,
        # explicitely declare this is not a CLI option
        extra_option: Annotated[bool, None] = False,
    ):
        assert my_option
        assert extra_option

    execute("my-session", posargs=["--my-option"])


def test_nox_parametrize(execute: Executor) -> None:
    group = NoxOpt()

    calls = []

    @group.session
    @parametrize("num", [1, 2, 3])
    def my_session(session: Session, num: int, mult: Annotated[int, Option()]) -> None:
        calls.append(num * mult)

    execute("my-session", posargs=["--mult", "2"])

    assert calls == [2, 4, 6]



from __future__ import annotations

import functools
from argparse import ArgumentParser
from typing import TYPE_CHECKING, Any, Callable, DefaultDict, TypeVar, overload

import nox
from nox.sessions import Session

from noxopt import _config
from noxopt._option import Option, get_function_options
from noxopt._tagging import AutoTag


if TYPE_CHECKING:
    from typing import Concatenate, ParamSpec

    AnyFunc = Callable[..., Any]

    F = TypeVar("F", bound=Callable[..., None])
    P = ParamSpec("P")
    R = TypeVar("R")

    def copy_method_signature(
        func: Callable[P, R]
    ) -> Callable[[AnyFunc], Callable[Concatenate[Any, P], R]]:
        ...

else:

    def copy_method_signature(func):
        return lambda f: f


class NoxOpt:
    """Define a group of NoxOpt group"""

    def __init__(
        self,
        parser: ArgumentParser | None = None,
        auto_tag: bool = False,
        where: dict[str, Any] | None = None,
        explicit_options: bool = False,
    ):
        self._parser = parser or ArgumentParser("")
        self._options_by_flags: dict[str, Option] = {}
        self._auto_tag = AutoTag() if auto_tag else None
        self._explicit_options = explicit_options
        self._setup_funcs: DefaultDict[
            str, list[Callable[[Session], None]]
        ] = DefaultDict(list)
        self._common_session_kwargs: dict[str, Any] = where or {}
        if _config.NOXOPT_TESTING:
            self._common_session_kwargs["venv_backend"] = "none"

    @copy_method_signature(nox.session)
    def session(
        self,
        func: Any | None = None,
        *args: Any,
        name: str | None = None,
        **kwargs: Any,
    ) -> Any:
        """Designate the decorated function as a session with command line arguments."""

        def decorator(main_func: Any) -> Any:
            session_name = name or main_func.__name__.replace("_", "-")
            wrapper = _copy_nox_parametrize(
                main_func,
                self._create_setup_wrapper(
                    session_name,
                    self._create_parser_wrapper(main_func),
                ),
            )
            session = nox.session(
                *args,
                name=session_name,
                **{**self._common_session_kwargs, **kwargs},
            )(wrapper)
            if self._auto_tag:
                self._auto_tag.add_func(session_name, session)
            return session

        return decorator if func is None else decorator(func)

    @overload
    def setup(self, __func: Callable[..., None]) -> Callable[..., None]:
        ...

    @overload
    def setup(self, *names: str) -> Callable[[AnyFunc], AnyFunc]:
        ...

    def setup(self, *args: Any) -> Any:
        """Define a function that will run before"""
        first_arg, *rest_args = args
        if isinstance(first_arg, str):
            prefixes = args
            func = None
        else:
            func = first_arg
            prefixes = ()

        def decorator(func: Callable[..., None]) -> Callable[..., None]:
            wrapper = self._create_parser_wrapper(func)
            for p in prefixes or [""]:
                self._setup_funcs[p].append(wrapper)
            return wrapper

        return decorator if func is None else decorator(func)

    def _create_setup_wrapper(self, name: str, func: AnyFunc) -> AnyFunc:
        @functools.wraps(func)
        def wrapper(session: Session, *args: Any, **kwargs: Any) -> None:
            for prefix, setup_funcs in self._setup_funcs.items():
                if name.startswith(prefix):
                    for f in setup_funcs:
                        f(session)
            func(session, *args, **kwargs)

        return wrapper

    def _create_parser_wrapper(self, func: AnyFunc) -> AnyFunc:
        own_params: set[str] = set()
        for name, option in get_function_options(func, self._explicit_options).items():
            own_params.add(name)
            self._add_option_to_parser(option)

        @functools.wraps(func)
        def wrapper(session: Session, *args: Any, **kwargs: Any) -> None:
            args_dict = self._parser.parse_args(session.posargs).__dict__
            func(
                session,
                *args,
                **kwargs,
                **{k: v for k, v in args_dict.items() if k in own_params},
            )

        return wrapper

    def _add_option_to_parser(self, option: Option) -> None:
        already_exists = False

        for flag in option.flags:
            existing = self._options_by_flags.get(flag)
            if not existing:
                self._options_by_flags[flag] = option
            elif existing != option:
                raise ValueError(
                    f"Conflicting session options:\n"
                    f"new:      {option}\n"
                    f"existing: {existing}"
                )
            else:
                already_exists = True

        if not already_exists:
            option.add_argument_to_parser(self._parser)


def _copy_nox_parametrize(func: AnyFunc, wrapper: AnyFunc) -> Callable[[Any], AnyFunc]:
    # nox.parametrize adds this attribbute
    if hasattr(func, "parametrize"):
        wrapper.parametrize = func.parametrize  # type: ignore[attr-defined]
    return wrapper



from __future__ import annotations

import sys
from argparse import ArgumentParser
from dataclasses import dataclass, fields, replace
from inspect import Parameter, signature
from typing import Any, Callable, Container, Sequence, cast


if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import Annotated, get_args, get_origin, get_type_hints
else:
    from typing import Annotated, get_args, get_origin, get_type_hints


UNDEFINED = cast(Any, type("Undefined", (), {"__repr__": lambda self: "UNDEFINED"})())


def get_function_options(
    func: Callable[..., Any], explicit_options: bool | None = None
) -> dict[str, Option]:
    options: dict[str, Option] = {}

    explicit_options = (
        explicit_options
        # we require parametrized sessions to have explicitely declared options
        or bool(getattr(func, "parametrize", None))
    )

    func = _get_wrapped_func(func)
    for k, (d, a) in _get_function_defaults_and_annotations(func).items():
        opt = _create_option(k, d, a, explicit_options)
        if opt is not None:
            options[k] = opt

    return options


@dataclass
class Option:
    flags: Sequence[str] = UNDEFINED
    # remainder are alphabetical
    action: str | None = UNDEFINED
    choices: Container[Any] = UNDEFINED
    const: Any = UNDEFINED
    default: Any | None = UNDEFINED
    dest: str = UNDEFINED
    help: str = UNDEFINED
    metavar: str = UNDEFINED
    nargs: str | int | None = UNDEFINED
    required: bool = UNDEFINED
    type: int | float | Callable[[Any], Any] = UNDEFINED

    def __post_init__(self) -> None:
        if isinstance(self.flags, str):
            self.flags = (self.flags,)
        if self.flags is not UNDEFINED:
            for f in self.flags:
                if not f.startswith("-"):
                    raise ValueError(f"Option only supports flags, but got {f!r}")

    def add_argument_to_parser(self, parser: ArgumentParser) -> None:
        kwargs = {
            k: v
            for k, v in (
                # Can't use asdict() since that deep copies and we need
                # to filter using an identity check against UNDEFINED.
                [(f.name, getattr(self, f.name)) for f in fields(self)]
            )
            if v is not UNDEFINED
        }

        flags = kwargs.pop("flags")
        parser.add_argument(*flags, **kwargs)


def _create_option(
    name: str, default: Any, annotation: Any, explicit_options: bool
) -> Option | None:
    if get_origin(annotation) is Annotated:
        anno_args = get_args(annotation)
    elif explicit_options:
        return None
    else:
        anno_args = (annotation, Option())

    opt: Option | None
    opt_type, opt, *extra_args = anno_args

    if extra_args:
        raise ValueError(f"{annotation} has extra metadata {extra_args}")

    if opt is None:
        return None

    if not isinstance(opt, Option):
        raise ValueError(f"{annotation} metadata must be an Option")

    if opt.type is UNDEFINED:
        if not callable(opt_type):
            raise TypeError(
                f"Annotation {annotation} for parameter {name!r} is not callable."
                f"Declare option type with Annotated[..., Option(type=...)] instead."
            )
        opt = replace(opt, type=opt_type)

    if opt.type is bool:
        opt = replace(
            opt,
            action="store_false" if default is True else "store_true",
            type=UNDEFINED,
            default=UNDEFINED,
        )
    elif default is not UNDEFINED:
        opt = replace(opt, default=default)
    else:
        opt = replace(opt, required=True)

    if opt.flags is UNDEFINED:
        opt = replace(opt, flags="--" + name.replace("_", "-"))

    return opt


def _get_function_defaults_and_annotations(
    func: Callable[..., Any]
) -> dict[str, tuple[Parameter, Any]]:
    parameters = list(signature(func).parameters.values())
    annotations = get_type_hints(func, include_extras=True)

    # skip first positional arg (the nox session obj)
    del parameters[0]

    kw_param_kinds = (Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY)
    non_kws = [p.name for p in parameters if p.kind not in kw_param_kinds]
    if non_kws:
        raise TypeError(
            f"Found non-keyword session parameters {', '.join(non_kws)} in {func}. If "
            "you used a decorator, it may not have preserved information about the "
            "original function using functools.wraps()."
        )

    return {
        param.name: (
            UNDEFINED if param.default is Parameter.empty else param.default,
            annotations[param.name],
        )
        for param in parameters
    }


def _get_wrapped_func(given_func: Callable[..., Any]) -> Callable[..., Any]:
    maybe_wrapped: Callable[..., Any] | None = given_func
    while maybe_wrapped is not None:
        func = maybe_wrapped
        maybe_wrapped = getattr(func, "__wrapped__", None)
    return func



from __future__ import annotations

import sys

from nox.sessions import Session

from noxopt._option import Option
from noxopt.noxopt import NoxOpt


if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import Annotated
else:
    from typing import Annotated

if sys.version_info < (3, 8):  # pragma: no cover
    from importlib_metadata import version as _version
else:
    from importlib.metadata import version as _version

__version__ = _version(__name__)
__all__ = ["NoxOpt", "Option", "Session", "Annotated"]



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    # avoid import error if this private module ever changes in the future
    from nox._decorators import Func


class AutoTag:
    r"""Session auto tagging utility

    Construct a graph of words, and at every branching point in the graph it creates a
    tag. For example the session names:

    - `a-x-1`
    - `a-x-2`
    - `a-y-1`
    - `a-y-2`
    - `b-x-1`
    - `b-x-2`

    Would create the graph:

    ```
          *
         / \
        a   b
       /\    \
      x  y    x
     /|  |\   |\
    1 2  1 2  1 2
    ```

    At every branching point in the graph a tag would be generated and applied to all
    functions with that prefix. So the tags would be:

    - `a`
    - `a-x`
    - `a-y`
    - `b-x`

    This isn't always perfect since technically, we could subtitute the tag `b-x` for
    just `b` since the shorted common string between `b-x-1` and `b-x-2` is `b`. But
    it works most of the time, and it makes the algorithm a bit easier to reason about.
    """

    def __init__(self, sep: str = "-"):
        self._sep = sep
        self._tag_tree = _TagNode()

    def add_func(self, name: str, func: Func) -> None:
        node = self._tag_tree

        # each branching point in the graph represents a tag
        for word in name.split(self._sep):
            if word in node.children:
                # this is a branching point in the graph
                if len(node.children) > 1:
                    node.add_tag(func)
                node = node.children[word]
            else:
                if len(node.children) == 1:
                    node.add_tag(func)
                    # We're about to create a new branching point - funcs added earlier
                    # will not have this nodes tag.
                    node.retroactively_add_tags()
                node = node.add_node(word, self._sep)

        # add this func to the last node
        node.add_func(func)


class _TagNode:
    def __init__(self, parent: _TagNode | None = None, tag: str | None = None):
        self.parent = parent
        self.tag = tag
        self.funcs: list[Func] = []
        self.children: dict[str, _TagNode] = {}

    def add_node(self, word: str, sep: str) -> _TagNode:
        tag = f"{self.tag}{sep}{word}" if self.tag else word
        child = self.children[word] = _TagNode(self, tag)
        return child

    def add_func(self, func: Func) -> None:
        self.funcs.append(func)

    def add_tag(self, func: Func) -> None:
        if self.tag:
            func.tags.append(self.tag)

    def retroactively_add_tags(self) -> None:
        to_visit = [self]
        while to_visit:
            node = to_visit.pop()
            for f in node.funcs:
                self.add_tag(f)
            to_visit.extend(node.children.values())
